'''
Instituto Tecnológico de Aeronáutica
CCI 36 – Computação Gráfica - 2018
Prof. Carlos Henrique Q. Forster
Luiz Felipe Simões Hoffmann

Exemplo 03 - Curva de Bézier
'''
from OpenGL.GL import *
from OpenGL.GL import shaders
import numpy as np
import os
import pygame as pg
import sys


def cubic_bezier(p, npoints):
    t = np.linspace(0., 1., npoints, dtype=np.float32).reshape((npoints, 1))
    bezier_points = (1 - t)**3 * p[:, 0] + \
                     3 * (1 - t)**2 * t * p[:, 1] + \
                     3 * (1 - t) * t**2 * p[:, 2] + \
                     t**3 * p[:, 3]
    return bezier_points.T


class OpenGLWindow:

    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec2 position;
layout (location = 1) in float color;
out float vertex_color;
void main()
{
    gl_Position = vec4(position, 0.0, 1.0);
    vertex_color = color;
}'''

    FRAGMENT_SHADER = '''#version 330 core
in float vertex_color;
out vec4 color;
void main()
{
    color = vec4(1.0f, vertex_color, 0.0f, 1.0f);
}'''
    
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        
        # Inicia o pygame
        pg.init()
        
        self.clock = pg.time.Clock()
        
        # Cria a janela
        pg.display.set_caption(title)
        screen = pg.display.set_mode((width, height),
                                     pg.HWSURFACE | pg.OPENGL | pg.DOUBLEBUF)
        self.resize(width, height)
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        self.control_points = np.array([[-.8,  .2, .4, .8],  # coordenada x
                                        [-.8, -.9, .9, .1]], # coordenada y
                              dtype=np.float32)
        self.npoints = 100
        bezier_points = cubic_bezier(self.control_points, self.npoints)
        
        points = np.concatenate((bezier_points.flatten('F'),
                                 self.control_points.flatten('F')))
        colors = np.ones((points.size // 2,), dtype=np.float32)
        colors[-self.control_points[0].size:] = 0.
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
        
        # Cria um vertex buffer object
        vbo_points = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo_points)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, points.size * points.itemsize, points,
                 GL_STATIC_DRAW)
        
        # Cria um vertex buffer object
        vbo_colors = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo_colors)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, colors.size * colors.itemsize, colors,
                     GL_STATIC_DRAW)
        
        # Especifica o local das coordenadas (posições) x, y
        glBindBuffer(GL_ARRAY_BUFFER, vbo_points)
        glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 2 * points.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Especifica o local das cores r, g, b
        glBindBuffer(GL_ARRAY_BUFFER, vbo_colors) 
        glVertexAttribPointer(1, 1, GL_FLOAT, GL_FALSE, colors.itemsize, None)
        glEnableVertexAttribArray(1)
        
        # Desconecta vao
        glBindVertexArray(0)
        
        # Wireframe
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
        
    def display(self):
        # Define a cor de fundo da janela
        glClearColor(0., 0., 0., 1.)
        
        # Limpa o buffer da janela
        glClear(GL_COLOR_BUFFER_BIT)
        
        # Desenha a curva de Bézier
        glUseProgram(self.program)
        glBindVertexArray(self.vao)
        glDrawArrays(GL_LINE_STRIP, 0, self.npoints)
        glPointSize(5)
        glDrawArrays(GL_POINTS, self.npoints, self.control_points[0].size)
        glDrawArrays(GL_LINE_STRIP, self.npoints, self.control_points[0].size)
        glBindVertexArray(0)
        glUseProgram(0)
    
    def resize(self, width, height):
        # Define o tamanho da janela de desenho
        glViewport(0, 0, width, height)
        
    def show(self):
        # Permanece em loop ate o usuário fechar a janela
        while True:
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    pg.quit()
                    sys.exit()
            
            #Renderiza
            self.display()
            
            # Atualiza a janela
            pg.display.flip()
            
            # Define a velocidade com que a janela é atualizada
            self.clock.tick(65)


def main():
    window = OpenGLWindow(800, 600, 'CCI-36 - Curva de Bézier')
    window.show()


if __name__ == '__main__':
    main()