'''
Instituto Tecnológico de Aeronáutica
CCI 36 – Computação Gráfica - 2018
Prof. Carlos Henrique Q. Forster
Luiz Felipe Simões Hoffmann

Exemplo 06 - Esfera
'''
from OpenGL.GL import *
from OpenGL.GL import shaders
import numpy as np
import os
import pygame as pg
import pyrr
import sys


class Sphere():
    def __init__(self, stacks=30, slices=30):
        self.stacks = stacks
        self.slices = slices
    
    def indices(self, dtype=np.uint32):
        indices = np.array([])
        v = 0
        for i in range(self.slices - 1):
            for j in range(self.stacks):
                next_ = (j + 1) % self.stacks
                indices = np.append(indices, [[v + j, v + next_, v + j + self.stacks]])
                indices = np.append(indices, [[v + next_, v + next_ + self.stacks, v + j + self.stacks]])
            v = v + self.stacks
        return indices.astype(dtype)
    
    def vertices(self, dtype=np.float32):
        u, v = np.meshgrid(np.linspace(0, 2 * np.pi, self.stacks),
                           np.linspace(0, np.pi, self.slices))
        
        x = (np.cos(u) * np.sin(v)).reshape((self.stacks * self.slices, 1))
        y = (np.sin(u) * np.sin(v)).reshape((self.stacks * self.slices, 1))
        z = (np.cos(v)).reshape((self.stacks * self.slices, 1))
        
        return np.concatenate((x, y, z), axis=1).flatten().astype(dtype)
    

class OpenGLWindow:

    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec3 position;
uniform mat4 view;
uniform mat4 model;
uniform mat4 projection;
void main()
{
    gl_Position = projection * view * model * vec4(position, 1.0);
}'''

    FRAGMENT_SHADER = '''#version 330 core
out vec4 color;
void main()
{
    color = vec4(1.0f, 0.5f, 0.2f, 1.0f);
}'''
    
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        
        # Inicia o pygame
        pg.init()
        
        self.clock = pg.time.Clock()
        
        # Cria a janela
        pg.display.set_caption(title)
        screen = pg.display.set_mode((width, height),
                                     pg.HWSURFACE | pg.OPENGL | pg.DOUBLEBUF)
        self.resize(width, height)
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        sphere = Sphere()
        vertices = sphere.vertices()
        self.indices = sphere.indices()
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
    
        # Cria um vertex buffer object
        vbo = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, vertices.size * vertices.itemsize,
                     vertices, GL_STATIC_DRAW)
        
        # Cria um element buffer object
        ebo = glGenBuffers(1)
        # Conecta objeto ebo
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
        # Copia dados na memória
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, self.indices.size * self.indices.itemsize,
                     self.indices, GL_STATIC_DRAW)
        
        # Especifica o local das coordenadas (posições) x, y, z
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * vertices.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Desconecta vao
        glBindVertexArray(0)
        
        # Cria matriz de modelo
        model = pyrr.Matrix44.from_x_rotation(np.radians(-45.), dtype=np.float32)
        
        # Cria matriz de projeção
        projection = pyrr.Matrix44.perspective_projection(45.,
                                                          self.width / self.height,
                                                          .1,
                                                          100.,
                                                          dtype=np.float32)
        
        glUseProgram(self.program)
        
        # Retorna a localização das variáveis uniformes
        model_loc = glGetUniformLocation(self.program, 'model')
        self.view_loc = glGetUniformLocation(self.program, 'view')
        proj_loc = glGetUniformLocation(self.program, 'projection')
        
        # Define o valor das variáveis uniformes
        glUniformMatrix4fv(model_loc, 1, GL_FALSE, model)
        glUniformMatrix4fv(proj_loc, 1, GL_FALSE, projection)
        
        glUseProgram(0)
        
        # Teste de profundidade (utilizar em 3D)
        glEnable(GL_DEPTH_TEST)
        
        # Wireframe
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
    
    def display(self):
        # Define a cor de fundo da janela
        glClearColor(0.2, 0.3, 0.3, 1.0)
        
        # Limpa o buffer da janela
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        glUseProgram(self.program)
        glBindVertexArray(self.vao)
        
        # Cria matriz de visualização
        radius = 4.0
        x = np.sin(pg.time.get_ticks() / 1000) * radius
        z = np.cos(pg.time.get_ticks() / 1000) * radius
        view = pyrr.Matrix44.look_at(pyrr.Vector3([x, 0., z], dtype=np.float16),
                                     pyrr.Vector3([0., 0., 0.], dtype=np.float16),
                                     pyrr.Vector3([0., 1., 0.], dtype=np.float16),
                                     dtype=np.float32)
        # Define o valor da variável uniforme `view`
        glUniformMatrix4fv(self.view_loc, 1, GL_FALSE, view)
        
        # Desenha a esfera
        glDrawElements(GL_TRIANGLES, self.indices.size, GL_UNSIGNED_INT, None)
        
        glBindVertexArray(0)
        glUseProgram(0)
    
    def resize(self, width, height):
        # Define o tamanho da janela de desenho
        glViewport(0, 0, width, height)
        
    def show(self):
        # Permanece em loop ate o usuário fechar a janela
        while True:
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    pg.quit()
                    sys.exit()
            
            #Renderiza
            self.display()
            
            # Atualiza a janela
            pg.display.flip()
            
            # Define a velocidade com que a janela é atualizada
            self.clock.tick(65)


def main():
    window = OpenGLWindow(800, 600, 'CCI-36 - Esfera')
    window.show()


if __name__ == '__main__':
    main()