'''
Instituto Tecnológico de Aeronáutica
CCI 36 – Computação Gráfica - 2018
Prof. Carlos Henrique Q. Forster
Luiz Felipe Simões Hoffmann

Exemplo 08 - Modelo de Iluminação Local de Phong
'''
from OpenGL.GL import *
from OpenGL.GL import shaders
import numpy as np
import os
import pygame as pg
import pyrr
import sys


class Sphere():
    def __init__(self, stacks=30, slices=30):
        self.stacks = stacks
        self.slices = slices
    
    def indices(self, dtype=np.uint32):
        indices = np.array([])
        v = 0
        for i in range(self.slices - 1):
            for j in range(self.stacks):
                next_ = (j + 1) % self.stacks
                indices = np.append(indices, [[v + j, v + next_, v + j + self.stacks]])
                indices = np.append(indices, [[v + next_, v + next_ + self.stacks, v + j + self.stacks]])
            v = v + self.stacks
        return indices.astype(dtype)
    
    def vertices(self, dtype=np.float32):
        u, v = np.meshgrid(np.linspace(0, 2 * np.pi, self.stacks),
                           np.linspace(0, np.pi, self.slices))
        
        x = (np.cos(u) * np.sin(v)).reshape((self.stacks * self.slices, 1))
        y = (np.sin(u) * np.sin(v)).reshape((self.stacks * self.slices, 1))
        z = (np.cos(v)).reshape((self.stacks * self.slices, 1))
        
        return np.concatenate((x, y, z), axis=1).flatten().astype(dtype)
    

class Light:
    
    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec3 position;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

void main()
{
    gl_Position = projection * view * model * vec4(position, 1.0);
}'''

    FRAGMENT_SHADER = '''#version 330 core
out vec4 color;

void main()
{
    color = vec4(1.0f);
}'''

    def __init__(self, view, projection):
        self.view = view
        self.projection = projection
        
        # Define a cor da luz
        color = pyrr.Vector3([2., .7, 1.3], dtype=np.float32)
        self.diffuse = .5 * color 
        self.ambient = .2 * self.diffuse
        self.specular = pyrr.Vector3([1., 1., 1.], dtype=np.float32)
        
        #self.position = pyrr.Vector3([.4, .95, 1.2], dtype=np.float16)
        self.position = pyrr.Vector3([1.2, .5, 1.4], dtype=np.float16)
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        sphere = Sphere()
        vertices = sphere.vertices()
        self.indices = sphere.indices()
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
    
        # Cria um vertex buffer object
        vbo = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, vertices.size * vertices.itemsize,
                     vertices, GL_STATIC_DRAW)
        
        # Cria um element buffer object
        ebo = glGenBuffers(1)
        # Conecta objeto ebo
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
        # Copia dados na memória
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, self.indices.size * self.indices.itemsize,
                     self.indices, GL_STATIC_DRAW)
        
        # Especifica o local das coordenadas (posições) x, y, z
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * vertices.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Desconecta objeto vao
        glBindVertexArray(0)
        
        # Cria a matriz de modelo
        model = pyrr.Matrix44.from_translation(self.position, dtype=np.float16)
        model *= pyrr.Matrix44.from_scale(pyrr.Vector3([.12, .12, .12], dtype=np.float16), dtype=np.float16)
        model = model.astype(np.float32)
        
        glUseProgram(self.program)
        
        # Retorna a localização das variáveis uniformes
        model_loc = glGetUniformLocation(self.program, 'model')
        view_loc = glGetUniformLocation(self.program, 'view')
        proj_loc = glGetUniformLocation(self.program, 'projection')
        
        # Define o valor das variáveis uniformes
        glUniformMatrix4fv(view_loc, 1, GL_FALSE, self.view)
        glUniformMatrix4fv(model_loc, 1, GL_FALSE, model)
        glUniformMatrix4fv(proj_loc, 1, GL_FALSE, self.projection)
        
        glUseProgram(0)


class Cube:
    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec3 position;
layout (location = 1) in vec3 normal;

out vec3 Normal;
out vec3 frag_pos;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

void main()
{
    gl_Position = projection * view * model * vec4(position, 1.0);
    Normal = mat3(transpose(inverse(model))) * normal;
    frag_pos = vec3(model * vec4(position, 1.0f));
}'''

    FRAGMENT_SHADER = '''#version 330 core
struct Material {
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
    float shininess;
};

struct Light {
    vec3 position;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

in vec3 Normal;
in vec3 frag_pos;

out vec4 color;

uniform Material material;
uniform Light light;
uniform vec3 camera_pos;

void main()
{
    // Ambiente
    vec3 ambient = light.ambient * material.ambient;
    
    // Difusa
    vec3 norm = normalize(Normal);
    vec3 light_dir = normalize(light.position - frag_pos);
    float diff = max(dot(norm, light_dir), 0.0);
    vec3 diffuse = light.diffuse * diff * material.diffuse;
    
    // Especular
    vec3 camera_dir = normalize(camera_pos - frag_pos);
    vec3 reflect_dir = reflect(-light_dir, norm);  
    float spec = pow(max(dot(camera_dir, reflect_dir), 0.0), material.shininess);
    vec3 specular = light.specular * spec * material.specular;
    
    color = vec4(ambient + diffuse + specular, 1.0f);
}'''

    def __init__(self, view, projection, light, camera_pos):
        self.view = view
        self.projection = projection
        self.light = light
        self.camera_pos = camera_pos
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        vertices = np.array([
                             -0.5, -0.5, -0.5,  0.0,  0.0, -1.0,
                              0.5, -0.5, -0.5,  0.0,  0.0, -1.0, 
                              0.5,  0.5, -0.5,  0.0,  0.0, -1.0, 
                              0.5,  0.5, -0.5,  0.0,  0.0, -1.0, 
                             -0.5,  0.5, -0.5,  0.0,  0.0, -1.0, 
                             -0.5, -0.5, -0.5,  0.0,  0.0, -1.0,
                             
                             -0.5, -0.5,  0.5,  0.0,  0.0, 1.0,
                              0.5, -0.5,  0.5,  0.0,  0.0, 1.0,
                              0.5,  0.5,  0.5,  0.0,  0.0, 1.0,
                              0.5,  0.5,  0.5,  0.0,  0.0, 1.0,
                             -0.5,  0.5,  0.5,  0.0,  0.0, 1.0,
                             -0.5, -0.5,  0.5,  0.0,  0.0, 1.0,
                             
                             -0.5,  0.5,  0.5, -1.0,  0.0,  0.0,
                             -0.5,  0.5, -0.5, -1.0,  0.0,  0.0,
                             -0.5, -0.5, -0.5, -1.0,  0.0,  0.0,
                             -0.5, -0.5, -0.5, -1.0,  0.0,  0.0,
                             -0.5, -0.5,  0.5, -1.0,  0.0,  0.0,
                             -0.5,  0.5,  0.5, -1.0,  0.0,  0.0,
                             
                              0.5,  0.5,  0.5,  1.0,  0.0,  0.0,
                              0.5,  0.5, -0.5,  1.0,  0.0,  0.0,
                              0.5, -0.5, -0.5,  1.0,  0.0,  0.0,
                              0.5, -0.5, -0.5,  1.0,  0.0,  0.0,
                              0.5, -0.5,  0.5,  1.0,  0.0,  0.0,
                              0.5,  0.5,  0.5,  1.0,  0.0,  0.0,
                              
                             -0.5, -0.5, -0.5,  0.0, -1.0,  0.0,
                              0.5, -0.5, -0.5,  0.0, -1.0,  0.0,
                              0.5, -0.5,  0.5,  0.0, -1.0,  0.0,
                              0.5, -0.5,  0.5,  0.0, -1.0,  0.0,
                             -0.5, -0.5,  0.5,  0.0, -1.0,  0.0,
                             -0.5, -0.5, -0.5,  0.0, -1.0,  0.0,
                             
                             -0.5,  0.5, -0.5,  0.0,  1.0,  0.0,
                              0.5,  0.5, -0.5,  0.0,  1.0,  0.0,
                              0.5,  0.5,  0.5,  0.0,  1.0,  0.0,
                              0.5,  0.5,  0.5,  0.0,  1.0,  0.0,
                             -0.5,  0.5,  0.5,  0.0,  1.0,  0.0,
                             -0.5,  0.5, -0.5,  0.0,  1.0,  0.0
                            ], dtype=np.float32)
        
        '''
        #                    posicao            normal
        vertices = np.array([-0.5, -0.5,  0.5,  0.0,  0.0,  1.0,
                              0.5, -0.5,  0.5,  0.0,  0.0,  1.0,
                              0.5,  0.5,  0.5,  0.0,  0.0,  1.0,
                             -0.5,  0.5,  0.5,  0.0,  0.0,  1.0,
                             
                             -0.5, -0.5, -0.5,  0.0,  0.0, -1.0,
                              0.5, -0.5, -0.5,  0.0,  0.0, -1.0,
                              0.5,  0.5, -0.5,  0.0,  0.0, -1.0,
                             -0.5,  0.5, -0.5,  0.0,  0.0, -1.0,
                             
                              0.5, -0.5, -0.5,  1.0,  0.0,  0.0,
                              0.5,  0.5, -0.5,  1.0,  0.0,  0.0,
                              0.5,  0.5,  0.5,  1.0,  0.0,  0.0,
                              0.5, -0.5,  0.5,  1.0,  0.0,  0.0,
                             
                             -0.5,  0.5, -0.5, -1.0,  0.0,  0.0,
                             -0.5, -0.5, -0.5, -1.0,  0.0,  0.0,
                             -0.5, -0.5,  0.5, -1.0,  0.0,  0.0,
                             -0.5,  0.5,  0.5, -1.0,  0.0,  0.0,
                            
                             -0.5, -0.5, -0.5,  0.0, -1.0,  0.0,
                              0.5, -0.5, -0.5,  0.0, -1.0,  0.0,
                              0.5, -0.5,  0.5,  0.0, -1.0,  0.0,
                             -0.5, -0.5,  0.5,  0.0, -1.0,  0.0,
                            
                              0.5,  0.5, -0.5,  0.0,  1.0,  0.0,
                             -0.5,  0.5, -0.5,  0.0,  1.0,  0.0,
                             -0.5,  0.5,  0.5,  0.0,  1.0,  0.0,
                              0.5,  0.5,  0.5,  0.0,  1.0,  0.0,
                            ], dtype=np.float32)
        
        self.indices = np.array([ 0,  1,  2,  2,  3,  0,
                                  4,  5,  6,  6,  7,  4,
                                  8,  9, 10, 10, 11,  8,
                                 12, 13, 14, 14, 15, 12,
                                 16, 17, 18, 18, 19, 16,
                                 20, 21, 22, 22, 23, 20
                                ], dtype=np.uint8)
        
        '''
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
    
        # Cria um vertex buffer object
        vbo = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, vertices.size * vertices.itemsize,
                     vertices, GL_STATIC_DRAW)
        
        '''
        # Cria um element buffer object
        ebo = glGenBuffers(1)
        # Conecta objeto ebo
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
        # Copia dados na memória
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, self.indices.size * self.indices.itemsize,
                     self.indices, GL_STATIC_DRAW)
        '''
        
        # Especifica o local das coordenadas (posições) x, y, z
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * vertices.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Especifica o local dos vetores normais 
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 6 * vertices.itemsize,
                              ctypes.c_void_p(3 * vertices.itemsize))
        glEnableVertexAttribArray(1)
        
        # Desconecta objeto vao
        glBindVertexArray(0)
        
        # Cria a matriz de modelo
        model = pyrr.Matrix44.from_scale(scale=(1.7, 1.7, 1.7), dtype=np.float16)
        model *= pyrr.Matrix44.from_x_rotation(np.radians(-45.), dtype=np.float16)
        model *= pyrr.Matrix44.from_y_rotation(np.radians(-45.), dtype=np.float16)
        model = model.astype(np.float32)
        
        glUseProgram(self.program)
        
        # Define o valor das variáveis uniformes
        glUniform3f(glGetUniformLocation(self.program, 'material.ambient'), 1., .5, .31)
        glUniform3f(glGetUniformLocation(self.program, 'material.diffuse'), 1., .5, 1.)
        glUniform3f(glGetUniformLocation(self.program, 'material.specular'), .5, .5, .5)
        glUniform1f(glGetUniformLocation(self.program, 'material.shininess'), 256.)
        
        glUniform3fv(glGetUniformLocation(self.program, 'light.position'), 1, self.light.position)
        glUniform3fv(glGetUniformLocation(self.program, 'light.ambient'), 1, self.light.ambient)
        glUniform3fv(glGetUniformLocation(self.program, 'light.diffuse'), 1, self.light.diffuse)
        glUniform3fv(glGetUniformLocation(self.program, 'light.specular'), 1, self.light.specular)
        
        glUniform3fv(glGetUniformLocation(self.program, 'camera_pos'), 1, self.camera_pos)
        
        glUniformMatrix4fv(glGetUniformLocation(self.program, 'model'), 1, GL_FALSE, model)
        glUniformMatrix4fv(glGetUniformLocation(self.program, 'view'), 1, GL_FALSE, self.view)
        glUniformMatrix4fv(glGetUniformLocation(self.program, 'projection'), 1, GL_FALSE, self.projection)
        
        glUseProgram(0)


class OpenGLWindow:
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        
        # Inicia o pygame
        pg.init()
        
        self.clock = pg.time.Clock()
        
        # Cria a janela
        pg.display.set_caption(title)
        screen = pg.display.set_mode((width, height),
                                     pg.HWSURFACE | pg.OPENGL | pg.DOUBLEBUF)
        self.resize(width, height)
        
        self.init()
    
    def init(self):
        # Cria a matriz de visualização
        camera = {'position': pyrr.Vector3([0., 0., 4.], dtype=np.float16),
                  'front': pyrr.Vector3([0., 0., -1.], dtype=np.float16),
                  'up': pyrr.Vector3([0., 1., 0.], dtype=np.float16)}
        view = pyrr.Matrix44.look_at(camera['position'],
                                     camera['position'] + camera['front'],
                                     camera['up'],
                                     dtype=np.float32)
        
        # Cria a matriz de projeção
        projection = pyrr.Matrix44.perspective_projection(45.,
                                                          self.width / self.height,
                                                          .1,
                                                          100.,
                                                          dtype=np.float32)
        
        self.light = Light(view, projection)
        self.cube = Cube(view, projection, self.light, camera['position'])
        
        # Teste de profundidade (utilizar em 3D)
        glEnable(GL_DEPTH_TEST)
        
    def display(self):
        # Define a cor de fundo da janela
        glClearColor(.1, .1, .1, 1.)
        
        # Limpa o buffer da janela
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        # Desenha a fonte de iluminação
        glUseProgram(self.light.program)
        glBindVertexArray(self.light.vao)
        glDrawElements(GL_TRIANGLES, self.light.indices.size, GL_UNSIGNED_INT, None)
        glBindVertexArray(0)
        glUseProgram(0)
        
        # Desenha o cubo
        glUseProgram(self.cube.program)
        glBindVertexArray(self.cube.vao)
        glDrawArrays(GL_TRIANGLES, 0, 36)
        #glDrawElements(GL_TRIANGLES, self.cube.indices.size, GL_UNSIGNED_BYTE, None)
        glBindVertexArray(0)
        glUseProgram(0)
        
    def resize(self, width, height):
        # Define o tamanho da janela de desenho
        glViewport(0, 0, width, height)
        
    def show(self):
        # Permanece em loop ate o usuário fechar a janela
        while True:
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    pg.quit()
                    sys.exit()
            
            #Renderiza
            self.display()
            
            # Atualiza a janela
            pg.display.flip()
            
            # Define a velocidade com que a janela é atualizada
            self.clock.tick(65)


def main():
    window = OpenGLWindow(800, 600, 'CCI-36 - Modelo de Iluminação Local de Phong')
    window.show()


if __name__ == '__main__':
    main()