'''
Instituto Tecnológico de Aeronáutica
CCI 36 – Computação Gráfica - 2018
Prof. Carlos Henrique Q. Forster
Luiz Felipe Simões Hoffmann

Exemplo 09 - Storm Trooper
'''
from objloader import ObjLoader
from OpenGL.GL import *
from OpenGL.GL import shaders
from PIL import Image
import numpy as np
import os
import pygame as pg
import pyrr
import sys


class Light:
    def __init__(self, view, projection):
        # Define a cor da luz
        color = pyrr.Vector3([2., .7, 1.3], dtype=np.float32)
        self.diffuse = .5 * color 
        self.ambient = .2 * self.diffuse
        self.specular = pyrr.Vector3([1., 1., 1.], dtype=np.float32)
        
        #self.position = pyrr.Vector3([.4, .95, 1.2], dtype=np.float16)
        self.position = pyrr.Vector3([1.2, .5, 1.4], dtype=np.float16)
        
        self.init()
    
    def init(self):
        pass


class Obj:
    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec3 position;
layout (location = 1) in vec2 texcoord;
layout (location = 2) in vec3 normal;

out vec2 Texcoord;
out vec3 Normal;
out vec3 frag_pos;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

void main()
{
    gl_Position = projection * view * model * vec4(position, 1.0);
    Texcoord = texcoord;
    Normal = mat3(transpose(inverse(model))) * normal;
    frag_pos = vec3(model * vec4(position, 1.0f));
}'''

    FRAGMENT_SHADER = '''#version 330 core
struct Material {
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
    float shininess;
};

struct Light {
    vec3 position;
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
};

in vec2 Texcoord;
in vec3 Normal;
in vec3 frag_pos;

out vec4 color;

uniform sampler2D tex;
uniform Material material;
uniform Light light;
uniform vec3 camera_pos;

void main()
{
    // Ambiente
    vec3 ambient = light.ambient * material.ambient;
    
    // Difusa
    vec3 norm = normalize(Normal);
    vec3 light_dir = normalize(light.position - frag_pos);
    float diff = max(dot(norm, light_dir), 0.0);
    vec3 diffuse = light.diffuse * diff * material.diffuse;
    
    // Especular
    vec3 camera_dir = normalize(camera_pos - frag_pos);
    vec3 reflect_dir = reflect(-light_dir, norm);  
    float spec = pow(max(dot(camera_dir, reflect_dir), 0.0), material.shininess);
    vec3 specular = light.specular * spec * material.specular;
    
    color = vec4(vec3(texture(tex, Texcoord)) * (ambient + diffuse + specular), 1.0f);
}'''

    def __init__(self, view, projection, light, camera_pos):
        self.view = view
        self.projection = projection
        self.light = light
        self.camera_pos = camera_pos
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        obj = ObjLoader()
        obj.load_model('obj/Stormtrooper.obj')
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
    
        # Cria o vertex buffer object das coordenadas (posições)
        self.vertices = np.array(obj.model_vertices, dtype=np.float32)
        vbo_vertices = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo_vertices)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, self.vertices.size * self.vertices.itemsize,
                     self.vertices, GL_STATIC_DRAW)
        
        # Especifica o local das coordenadas (posições) x, y, z
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * self.vertices.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Cria o vertex buffer object das texturas
        textures = np.array(obj.model_textures, dtype=np.float32)
        vbo_normals = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo_normals)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, textures.size * textures.itemsize,
                     textures, GL_STATIC_DRAW)
        
        # Especifica o local dos vetores normais 
        glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 2 * textures.itemsize, None)
        glEnableVertexAttribArray(1)
        
        # Cria um texture object
        texture = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, texture)
        
        # Definir parâmetros de repetição de padrões (wrapping)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT)
        
        # Definir parâmetros de filtragem
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        
        # Carregar imagem e vincular textura
        image = Image.open('img/Stormtrooper.png')
        img_data = np.array(list(image.getdata()), np.uint8)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, image.width, image.height, 0, GL_RGBA, GL_UNSIGNED_BYTE, img_data)
        glEnable(GL_TEXTURE_2D)
        
        # Cria o vertex buffer object dos vetores normais
        normals = np.array(obj.model_normals, dtype=np.float32)
        vbo_normals = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo_normals)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, normals.size * normals.itemsize,
                     normals, GL_STATIC_DRAW)
        
        # Especifica o local dos vetores normais 
        glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 3 * normals.itemsize, None)
        glEnableVertexAttribArray(2)
        
        # Desconecta objeto vao
        glBindVertexArray(0)
        
        # Cria a matriz de modelo
        model = pyrr.Matrix44.from_translation(pyrr.Vector3([0., -1.25, 0.], dtype=np.float32), dtype=np.float16)
        model *= pyrr.Matrix44.from_scale(scale=(.7, .7, .7), dtype=np.float16)
        #model *= pyrr.Matrix44.from_x_rotation(np.radians(-45.), dtype=np.float16)
        #model *= pyrr.Matrix44.from_y_rotation(np.radians(-45.), dtype=np.float16)
        model = model.astype(np.float32)
        
        glUseProgram(self.program)
        
        # Define o valor das variáveis uniformes
        glUniform3f(glGetUniformLocation(self.program, 'material.ambient'), 1., .5, .31)
        glUniform3f(glGetUniformLocation(self.program, 'material.diffuse'), 1., .5, 1.)
        glUniform3f(glGetUniformLocation(self.program, 'material.specular'), .5, .5, .5)
        glUniform1f(glGetUniformLocation(self.program, 'material.shininess'), 32.)
        
        glUniform3fv(glGetUniformLocation(self.program, 'light.position'), 1, self.light.position)
        glUniform3fv(glGetUniformLocation(self.program, 'light.ambient'), 1, self.light.ambient)
        glUniform3fv(glGetUniformLocation(self.program, 'light.diffuse'), 1, self.light.diffuse)
        glUniform3fv(glGetUniformLocation(self.program, 'light.specular'), 1, self.light.specular)
        
        glUniform3fv(glGetUniformLocation(self.program, 'camera_pos'), 1, self.camera_pos)
        
        glUniformMatrix4fv(glGetUniformLocation(self.program, 'model'), 1, GL_FALSE, model)
        #glUniformMatrix4fv(glGetUniformLocation(self.program, 'view'), 1, GL_FALSE, self.view)
        glUniformMatrix4fv(glGetUniformLocation(self.program, 'projection'), 1, GL_FALSE, self.projection)
        
        glUseProgram(0)
    
    def update_view_model(self):
        # Cria matriz de visualização
        radius = 4.0
        x = np.sin(pg.time.get_ticks() / 1000) * radius
        z = np.cos(pg.time.get_ticks() / 1000) * radius
        view = pyrr.Matrix44.look_at(pyrr.Vector3([x, 0., z], dtype=np.float16),
                                     pyrr.Vector3([0., 0., 0.], dtype=np.float16),
                                     pyrr.Vector3([0., 1., 0.], dtype=np.float16),
                                     dtype=np.float32)
        # Define o valor da variável uniforme `view`
        glUniformMatrix4fv(glGetUniformLocation(self.program, 'view'), 1, GL_FALSE, view)


class OpenGLWindow:
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        
        # Inicia o pygame
        pg.init()
        
        self.clock = pg.time.Clock()
        
        # Cria a janela
        pg.display.set_caption(title)
        screen = pg.display.set_mode((width, height),
                                     pg.HWSURFACE | pg.OPENGL | pg.DOUBLEBUF)
        self.resize(width, height)
        
        self.init()
    
    def init(self):
        # Cria a matriz de visualização
        camera = {'position': pyrr.Vector3([0., 0., 4.], dtype=np.float16),
                  'front': pyrr.Vector3([0., 0., -1.], dtype=np.float16),
                  'up': pyrr.Vector3([0., 1., 0.], dtype=np.float16)}
        view = pyrr.Matrix44.look_at(camera['position'],
                                     camera['position'] + camera['front'],
                                     camera['up'],
                                     dtype=np.float32)
        
        # Cria a matriz de projeção
        projection = pyrr.Matrix44.perspective_projection(45.,
                                                          self.width / self.height,
                                                          .1,
                                                          100.,
                                                          dtype=np.float32)
        
        self.light = Light(view, projection)
        self.obj = Obj(view, projection, self.light, camera['position'])
        
        # Teste de profundidade (utilizar em 3D)
        glEnable(GL_DEPTH_TEST)
        
    def display(self):
        # Define a cor de fundo da janela
        glClearColor(.1, .1, .1, 1.)
        
        # Limpa o buffer da janela
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        # Desenha o objeto
        glUseProgram(self.obj.program)
        self.obj.update_view_model()
        glBindVertexArray(self.obj.vao)
        glDrawArrays(GL_TRIANGLES, 0, self.obj.vertices.size // 3)
        #glDrawElements(GL_TRIANGLES, self.cube.indices.size, GL_UNSIGNED_BYTE, None)
        glBindVertexArray(0)
        glUseProgram(0)
        
    def resize(self, width, height):
        # Define o tamanho da janela de desenho
        glViewport(0, 0, width, height)
        
    def show(self):
        # Permanece em loop ate o usuário fechar a janela
        while True:
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    pg.quit()
                    sys.exit()
            
            #Renderiza
            self.display()
            
            # Atualiza a janela
            pg.display.flip()
            
            # Define a velocidade com que a janela é atualizada
            self.clock.tick(65)


def main():
    window = OpenGLWindow(800, 600, 'CCI-36 - Storm Trooper')
    window.show()


if __name__ == '__main__':
    main()